r"""The streamlit-passwordless package.

streamlit-passwordless provides a user model for Streamlit applications based on the Bitwarden
passwordless technology. It allows users to securely authenticate with a Streamlit application
using the passkey FIDO2 and WebAuthn protocols.
"""

# Standard library
from datetime import date

# Local
from streamlit_passwordless.bitwarden_passwordless import (
    BitwardenPasswordlessClient,
    BitwardenPasswordlessVerifiedUser,
    BitwardenRegisterConfig,
    register_button,
    sign_in_button,
)
from streamlit_passwordless.components import (
    ICON_ERROR,
    ICON_SUCCESS,
    ICON_WARNING,
    bitwarden_register_form,
    bitwarden_sign_in_form,
    init_session_state,
)
from streamlit_passwordless.exceptions import (
    RegisterUserError,
    SignInTokenVerificationError,
    StreamlitPasswordlessError,
)

__versiontuple__ = (0, 5, 1)
r"""The version of streamlit-passwordless in a comparable form.
Adheres to `Semantic Versioning <https://semver.org/spec/v2.0.0.html>`_
(MAJOR.MINOR.PATCH).
"""

__version__ = '0.5.1'
r"""The streamlit-passwordless version string."""

__releasedate__ = date(2024, 5, 25)
r"""The release date of the version specified in :data:`__versiontuple__`."""


# The Public API

__all__ = [
    # bitwarden_passwordless
    'BitwardenPasswordlessClient',
    'BitwardenPasswordlessVerifiedUser',
    'BitwardenRegisterConfig',
    'register_button',
    'sign_in_button',
    # components
    'ICON_ERROR',
    'ICON_SUCCESS',
    'ICON_WARNING',
    'bitwarden_register_form',
    'bitwarden_sign_in_form',
    'init_session_state',
    # exceptions
    'RegisterUserError,',
    'SignInTokenVerificationError',
    'StreamlitPasswordlessError',
    # attributes
    '__versiontuple__',
    '__version__',
    '__releasedate__',
]



r"""The sign in form component."""

# Standard library
import logging
from typing import Literal

# Third party
import streamlit as st

from streamlit_passwordless import exceptions
from streamlit_passwordless.bitwarden_passwordless.client import BitwardenPasswordlessClient
from streamlit_passwordless.bitwarden_passwordless.frontend import sign_in_button

# Local
from . import config, ids

logger = logging.getLogger(__name__)


def bitwarden_sign_in_form(
    client: BitwardenPasswordlessClient,
    with_alias: bool = True,
    with_discoverable: bool = True,
    with_autofill: bool = False,
    title: str = '#### Sign in',
    border: bool = True,
    submit_button_label: str = 'Sign in',
    button_type: Literal['primary', 'secondary'] = 'primary',
    alias_label: str = 'Alias',
    alias_max_length: int | None = 50,
    alias_placeholder: str | None = 'john.doe@example.com',
    alias_help: str | None = '__default__',
) -> None:
    r"""Render the Bitwarden Passwordless sign in form.

    Allows the user to sign in to the application with a registered passkey.

    Parameters
    ----------
    client : streamlit_passwordless.BitwardenPasswordlessClient
        The Bitwarden Passwordless client to use for interacting with
        the Bitwarden Passwordless application.

    with_alias : bool, default True
        If True the field to enter the alias to use for signing in will be rendered. If False
        the field is not rendered and the other sign in methods `with_discoverable` or
        `with_autofill` should be used. If the user specifies an alias it will override the other
        sign in methods `with_discoverable` and `with_autofill`.

    with_discoverable : bool, default True
        If True the browser's native UI prompt will be used to select the passkey to use for
        signing in. If False the sign in method is disabled. If `alias` is specified it will
        override this sign in method.

    with_autofill : bool, default False
        If True the browser's native autofill UI will be used to select the passkey to use for
        signing in. If False the sign in method is disabled. This method of signing in is
        overridden if `alias` specified or `with_discoverable` is True.

    title : str, default '#### Sign in.'
        The title of the sign in from. Markdown is supported.

    border : bool, default True
        True if a border surrounding the form should be rendered and False
        to remove the border.

    submit_button_label : str, default 'Sign in'
        The label of the submit button.

    button_type : Literal['primary', 'secondary'], default 'primary'
        The styling of the button. Emulates the `type` parameter of :func:`streamlit.button`.

    Other Parameters
    ----------------
    alias_label : str, default 'Alias'
        The label of the alias field.

    alias_max_length : int or None, default 50
        The maximum allowed number of characters of the alias field.
        If None the upper limit is removed.

    alias_placeholder : str or None, default 'john.doe@example.com'
        The placeholder of the alias field. If None the placeholder is removed.

    alias_help : str or None, default '__default__'
        The help text to display for the alias field. If '__default__' a sensible default
        help text will be used and if None the help text is removed.
    """

    error_msg = ''
    help: str | None = None
    banner_container = st.empty()

    if with_alias is False and with_discoverable is False and with_autofill is False:
        with banner_container:
            error_msg = (
                f'At least one sign in method must be chosen!\n'
                f'*{with_alias=}*, *{with_discoverable=}*,  *{with_autofill=}*'
            )
            logger.error(error_msg)
            st.error(error_msg, icon=config.ICON_ERROR)
            return

    with st.container(border=border):
        st.markdown(title)

        if with_alias:
            if alias_help == '__default__':
                help = (
                    'An alias of the user. If not supplied auto discover '
                    'of available credentials will be attempted.'
                )
            else:
                help = alias_help

            st.text_input(
                label=alias_label,
                placeholder=alias_placeholder,
                max_chars=alias_max_length,
                help=help,
                key=ids.BP_SIGN_IN_FORM_ALIAS_TEXT_INPUT,
            )

        token, error, clicked = sign_in_button(
            public_key=client.public_key,
            alias=st.session_state.get(ids.BP_SIGN_IN_FORM_ALIAS_TEXT_INPUT),
            with_discoverable=with_discoverable,
            with_autofill=with_autofill,
            label=submit_button_label,
            button_type=button_type,
            key=ids.BP_SIGN_IN_FORM_SUBMIT_BUTTON,
        )

    if not clicked:
        return

    if not token and error:
        error_msg = f'Error signing in!\nerror : {error}'
        logger.error(error_msg)
        with banner_container:
            st.error(error_msg, icon=config.ICON_ERROR)
        return

    try:
        verified_user = client.verify_sign_in(token=token)
    except exceptions.SignInTokenVerificationError as e:
        error_msg = str(e)
        logger.error(error_msg)
    except exceptions.StreamlitPasswordlessError as e:
        error_msg = f'Error creating verified user!\n{str(e)}'
        logger.error(error_msg)

    if error_msg:
        with banner_container:
            st.error(error_msg, icon=config.ICON_ERROR)
        return

    st.session_state[config.SK_BP_VERIFIED_USER] = verified_user
    with banner_container:
        st.success(
            f'Successfully signed in user {verified_user.credential_nickname}',
            icon=config.ICON_SUCCESS,
        )



r"""Functions that are common to several modules."""

# Standard library
from datetime import datetime
import logging
from zoneinfo import ZoneInfo

# Third party

# Local


logger = logging.getLogger(__name__)


def get_current_datetime(tz='UTC') -> datetime:
    r"""Get the current datetime based on the timezone `tz`."""

    return datetime.now(tz=ZoneInfo(tz))



r"""The exception hierarchy of streamlit-passwordless."""

# Standard library
from typing import Any


class StreamlitPasswordlessError(Exception):
    r"""The base Exception of streamlit-passwordless.

    Parameters
    ----------
    message : str
        The exception message.

    data : Any, default None
        Optional extra data to include in the exception.
    """

    def __init__(self, message: str, data: Any = None) -> None:
        self.data = data
        super().__init__(message)


class RegisterUserError(StreamlitPasswordlessError):
    r"""Raised for errors when registering a new user."""


class SignInTokenVerificationError(StreamlitPasswordlessError):
    r"""Raised for errors when the backend is verifying the sign in token."""



r"""The functions and models for interacting with the Bitwarden Passwordless backend API."""

# Standard library
import logging
from datetime import datetime, timedelta
from typing import Literal, Self, TypeAlias

# Third party
from passwordless import (
    PasswordlessClient,
    PasswordlessClientBuilder,
    PasswordlessError,
    PasswordlessOptions,
    RegisterToken,
    VerifiedUser,
    VerifySignIn,
)
from pydantic import AnyHttpUrl

# Local
from streamlit_passwordless import common, exceptions, models

BackendClient: TypeAlias = PasswordlessClient

logger = logging.getLogger(__name__)


class BitwardenRegisterConfig(models.BaseModel):
    r"""The available passkey configuration when registering a new user.

    See the `Bitwarden Passwordless`_ documentation for more info about the parameters.

    .. _ Bitwarden Passwordless: https://docs.passwordless.dev/guide/api.html#register-token

    Parameters
    ----------
    attestation : Literal['none', 'direct', 'indirect'], default 'none'
        WebAuthn attestation conveyance preference. 'direct' and 'indirect' are exclusive to the
        Enterprise plan of Bitwarden Passwordless. Trial & Pro plans are limited to 'none'.

    authenticator_type : Literal['any', 'platform', 'cross-platform'], default 'any'
        WebAuthn authenticator attachment modality. 'platform' refers to platform specific options
        such as Windows Hello, FaceID or TouchID, while 'cross-platform' means roaming devices such
        as security keys. 'any' (default) means any authenticator type is allowed.

    discoverable : bool, default True
        True allows the user to sign in without a username or alias by creating a
        client-side discoverable credential.

    user_verification : Literal['preferred', 'required', 'discouraged'], default 'preferred'
        Set the preference for how user verification (e.g. PIN code or biometrics) works when
        authenticating.

    validity : timedelta, default timedelta(seconds=120)
        When the registration token expires and becomes invalid defined as an offset
        from the start of the registration process.

    alias_hashing : bool, default True
        True means that aliases for a user are hashed before they are stored in the
        Bitwarden Passwordless database.
    """

    attestation: Literal['none', 'direct', 'indirect'] = 'none'
    authenticator_type: Literal['any', 'platform', 'cross-platform'] = 'any'
    discoverable: bool = True
    user_verification: Literal['preferred', 'required', 'discouraged'] = 'preferred'
    validity: timedelta = timedelta(seconds=120)
    alias_hashing: bool = True

    @property
    def expires_at(self) -> datetime:
        r"""The expiry time of the registration token in timezone UTC."""

        return common.get_current_datetime() + self.validity


class BitwardenPasswordlessVerifiedUser(models.BaseModel):
    r"""A verified user from Bitwarden Passwordless.

    The model is generated after a successful sign in process with Bitwarden Passwordless.

    Parameters
    ----------
    success : bool
        True if the user was successfully signed in and False otherwise.

    user_id: str
        The unique ID of the user that signed in.

    sign_in_timestamp : datetime
        The timestamp in timezone UTC when the user signed in.

    origin : AnyHttpUrl
        The domain name or IP-address of the application the user signed in to.

    device : str
        The device the user signed in from. E.g. a web browser.

    country : str
        The country code of the country that the user signed in from, e.g. SE for Sweden.

    credential_nickname : str
        The nickname for the passkey that was used to sign in.

    credential_id : str
        The ID of the passkey credential used to sign in. Base64 encoded.

    expires_at : str
        ?

    token_id : str
        ?

    type : str
        The type of sign in method e.g. 'passkey_signin'.

    rp_id : str or None
        The ID of the relaying party, which is that is the server that
        verifies the credentials during the sign in process.
    """

    success: bool
    user_id: str
    sign_in_timestamp: datetime
    origin: AnyHttpUrl
    device: str
    country: str
    credential_nickname: str
    credential_id: str
    expires_at: datetime
    token_id: str
    type: str
    rp_id: str | None

    def __str__(self) -> str:
        return (
            f'{self.__class__.__name__}(success={self.success}, '
            f'user_id={self.user_id}, '
            f'sign_in_timestamp={self.sign_in_timestamp.strftime(r"%Y-%m-%d %H:%M:%S %z")})'
        )

    @classmethod
    def _from_passwordless_verified_user(cls, verified_user: VerifiedUser) -> Self:
        r"""Create an instance from a `passwordless.VerifiedUser`."""

        return cls(
            success=verified_user.success,
            user_id=verified_user.user_id,
            sign_in_timestamp=verified_user.timestamp,
            origin=verified_user.origin,
            device=verified_user.device,
            country=verified_user.country,
            credential_nickname=verified_user.nickname,
            credential_id=verified_user.credential_id,
            expires_at=verified_user.expires_at,
            token_id=verified_user.token_id,
            type=verified_user.type,
            rp_id=verified_user.rp_id,
        )


def _build_backend_client(private_key: str, url: str) -> BackendClient:
    r"""Build the Bitwarden Passwordless backend client.

    Parameters
    ----------
    private_key : str
        The private key that the client uses for authenticating with the
        Bitwarden Passwordless backend API.

    url : str
        The base url to the backend API.

    Returns
    -------
    BackendClient
        The backend client.
    """

    try:
        options = PasswordlessOptions(api_secret=private_key, api_url=url)
        return PasswordlessClientBuilder(options=options).build()
    except Exception as e:
        error_msg = f'Could not build Bitwarden backend client! {type(e).__name__} : {str(e)}'
        raise exceptions.StreamlitPasswordlessError(error_msg) from None


def _create_register_token(
    client: BackendClient, user: models.User, register_config: BitwardenRegisterConfig
) -> str:
    r"""Create a register token to use for registering a device for a user.

    Parameters
    ----------
    client : BackendClient
        The Bitwarden Passwordless backend client to use for creating the register token.

    user : streamlit_passwordless.models.User
        The user to register.

    register_config : BitwardenRegisterConfig
        The configuration for creating the register token.

    Returns
    -------
    str
        The token to use for registering a device for a user.

    Raises
    ------
    streamlit_passwordless.exceptions.RegisterUserError
        If an error occurs while trying to create the register token using the
        Bitwarden Passwordless backend API.
    """

    input_register_config = RegisterToken(
        user_id=user.user_id,
        username=user.username,
        display_name=user.displayname,
        attestation=register_config.attestation,
        authenticator_type=register_config.authenticator_type,
        discoverable=register_config.discoverable,
        user_verification=register_config.user_verification,
        aliases=user.aliases,
        alias_hashing=register_config.alias_hashing,
        expires_at=register_config.expires_at,
    )

    try:
        registered_token = client.register_token(register_token=input_register_config)
    except PasswordlessError as e:
        error_msg = f'Error creating register token! {str(e)}\nproblem_details: {e.problem_details}'
        data = {
            'input_register_config': input_register_config,
            'problem_details': e.problem_details,
        }
        logger.error(error_msg)
        raise exceptions.RegisterUserError(error_msg, data=data) from None
    else:
        logger.info(f'Successfully created register token for user_id={user.user_id}')

    return registered_token.token  # type: ignore


def _verify_sign_in_token(client: BackendClient, token: str) -> BitwardenPasswordlessVerifiedUser:
    r"""Verify the sign in token to complete the sign in process.

    The sign in token is generated from the `frontend._sign_in` function.

    Parameters
    ----------
    client : BackendClient
        The Bitwarden Passwordless backend client to communicate
        with the Bitwarden Passwordless backend.

    token : str
        The token to verify.

    Returns
    -------
    BitwardenPasswordlessVerifiedUser
        Details from Bitwarden Passwordless about the user that was signed in.

    Raises
    ------
    streamlit_passwordless.SignInTokenVerificationError
        If the `token` cannot be verified successfully.

    streamlit_passwordless.StreamlitPasswordlessError
        If an instance of `BitwardenPasswordlessVerifiedUser` cannot be successfully created.
    """

    try:
        _verified_user = client.sign_in(verify_sign_in=VerifySignIn(token=token))
    except PasswordlessError as e:
        error_msg = f'Error verifying the sign in token!\nproblem_details: {e.problem_details}'
        data = {
            'token': token,
            'problem_details': e.problem_details,
        }
        raise exceptions.SignInTokenVerificationError(error_msg, data=data) from None

    return BitwardenPasswordlessVerifiedUser._from_passwordless_verified_user(
        verified_user=_verified_user
    )



r"""The register-form component and its callback functions."""

# Standard library
import logging
from typing import Literal

# Third party
import streamlit as st

# Local
from streamlit_passwordless import exceptions, models
from streamlit_passwordless.bitwarden_passwordless.client import BitwardenPasswordlessClient
from streamlit_passwordless.bitwarden_passwordless.frontend import register_button

from . import config, ids

logger = logging.getLogger(__name__)


def _validate_username() -> None:
    r"""Validate the input username."""

    username = st.session_state[ids.BP_REGISTER_FORM_USERNAME_TEXT_INPUT]
    if not username:
        error_msg = 'The username field is required!'
        logger.error(error_msg)
        st.error(error_msg, icon=config.ICON_ERROR)


def _create_user(
    username: str, displayname: str | None = None, aliases: str | None = None
) -> tuple[models.User | None, str]:
    r"""Create a new user to register.

    Parameters
    ----------
    username : str
        The username.

    displayname : str or None, default None
        The optional displayname of the user.

    aliases : str | None, default None
        The optional aliases of the user as a semicolon separated string.

    Returns
    -------
    user : models.User or None
        The user to create. None is returned if an error occurred while creating the user.

    error_msg : str
        The error message if a user could not be created successfully. If no errors an empty
        string is returned.
    """

    error_msg = ''
    try:
        user = models.User(username=username, displayname=displayname, aliases=aliases)
    except exceptions.StreamlitPasswordlessError as e:
        error_msg = str(e)
        logger.error(error_msg)
        user = None
    else:
        logger.debug(f'Successfully created user: {user}')

    return user, error_msg


def bitwarden_register_form(
    client: BitwardenPasswordlessClient,
    is_admin: bool = False,
    pre_authorized: bool = True,
    with_displayname: bool = False,
    with_alias: bool = False,
    title: str = '#### Register a new passkey with your device',
    border: bool = True,
    submit_button_label: str = 'Register',
    button_type: Literal['primary', 'secondary'] = 'primary',
    username_label: str = 'Username',
    username_max_length: int | None = 50,
    username_placeholder: str | None = 'john.doe@example.com',
    username_help: str | None = '__default__',
    displayname_label: str = 'Displayname',
    displayname_max_length: int | None = 50,
    displayname_placeholder: str | None = 'John Doe',
    displayname_help: str | None = '__default__',
    alias_label: str = 'Alias',
    alias_max_length: int | None = 50,
    alias_placeholder: str | None = 'j;john;jd',
    alias_help: str | None = '__default__',
) -> None:
    r"""Render the Bitwarden Passwordless register form.

    Allows the user to register an account with the application by creating
    and registrering a passkey with the user's device.

    Parameters
    ----------
    client : streamlit_passwordless.BitwardenPasswordlessClient
        The Bitwarden Passwordless client to use for interacting with
        the Bitwarden Passwordless application.

    is_admin : bool, default False
        True means that the user will be registered as an admin.
        Not implemented yet.

    pre_authorized : bool, default True
        If True require the username to exist in the pre_authorized table of the database
        to allow the user to register. Not implemented yet.

    with_displayname : bool, default False
        If True the displayname field will be added to the form allowing
        the user to fill out a displayname for the account.

    with_alias : bool, default False
        If True the alias field will be added to the form allowing the user to fill
        out one or multiple aliases for the account. Aliases should be separated by
        by semicolon (";"). An alias must be unique across all users.

    title : str, default '#### Register a new passkey with your device'
        The title of the registration from. Markdown is supported.

    border : bool, default True
        If True a border will be rendered around the form.

    submit_button_label : str, default 'Register'
        The label of the submit button.

    button_type : Literal['primary', 'secondary'], default 'primary'
        The styling of the button. Emulates the `type` parameter of :func:`streamlit.button`.

    Other Parameters
    ----------------
    username_label : str, default 'Username'
        The label of the username field.

    username_max_length : int or None, default 50
        The maximum allowed number of characters of the username field.
        If None the upper limit is removed.

    username_placeholder : str or None, default 'john.doe@example.com'
        The placeholder of the username field. If None the placeholder is removed.

    username_help : str or None, default '__default__'
        The help text to display for the username field. If '__default__' a sensible default
        help text will be used and if None the help text is removed.

    displayname_label : str, default 'Displayname'
        The label of the displayname field.

    displayname_max_length : int or None, default 50
        The maximum allowed number of characters of the displayname field.
        If None the upper limit is removed.

    displayname_placeholder : str or None, default 'John Doe'
        The placeholder of the displayname field. If None the placeholder is removed.

    displayname_help : str or None, default '__default__'
        The help text to display for the displayname field. If '__default__' a sensible default
        help text will be used and if None the help text is removed.

    alias_label : str, default 'Alias'
        The label of the alias field.

    alias_max_length : int or None, default 50
        The maximum allowed number of characters of the alias field.
        If None the upper limit is removed.

    alias_placeholder : str or None, default 'j;john;jd'
        The placeholder of the alias field. If None the placeholder is removed.

    alias_help : str or None, default '__default__'
        The help text to display for the alias field. If '__default__' a sensible default
        help text will be used and if None the help text is removed.
    """

    user = None
    register_token = ''
    error_msg = ''
    use_default_help = '__default__'
    help: str | None = None
    banner_container = st.empty()

    with st.container(border=border):
        st.markdown(title)

        if username_help == use_default_help:
            help = 'A unique identifier for the account. E.g. an email address.'
        else:
            help = username_help

        username = st.text_input(
            label=username_label,
            placeholder=username_placeholder,
            max_chars=username_max_length,
            help=help,
            on_change=_validate_username,
            key=ids.BP_REGISTER_FORM_USERNAME_TEXT_INPUT,
        )
        disabled = False if username else True

        if with_displayname:
            if displayname_help == use_default_help:
                help = 'A descriptive name of the user.'
            else:
                help = displayname_help

            displayname = st.text_input(
                label=displayname_label,
                placeholder=displayname_placeholder,
                max_chars=displayname_max_length,
                help=help,
                disabled=disabled,
                key=ids.BP_REGISTER_FORM_DISPLAYNAME_TEXT_INPUT,
            )
        else:
            displayname = None

        if with_alias:
            if alias_help == use_default_help:
                help = (
                    'One or more aliases that can be used to sign in to the account. '
                    'Aliases are separated by semicolon (";"). The username is always '
                    'added as an alias. An alias must be unique across all users.'
                )
            else:
                help = alias_help

            aliases = st.text_input(
                label=alias_label,
                placeholder=alias_placeholder,
                max_chars=alias_max_length,
                help=help,
                disabled=disabled,
                key=ids.BP_REGISTER_FORM_ALIASES_TEXT_INPUT,
            )
        else:
            aliases = None

        if username:
            user, error_msg = _create_user(
                username=username, displayname=displayname, aliases=aliases
            )
            if user is not None:
                try:
                    register_token = client.create_register_token(user=user)
                except exceptions.RegisterUserError as e:
                    error_msg = str(e)
                    logger.error(error_msg)

        token, error, clicked = register_button(
            register_token=register_token,
            public_key=client.public_key,
            credential_nickname=username,
            disabled=disabled,
            label=submit_button_label,
            button_type=button_type,
            key=ids.BP_REGISTER_FORM_SUBMIT_BUTTON,
        )

    if disabled or not clicked:
        return

    if error_msg:
        with banner_container:
            st.error(error_msg, icon=config.ICON_ERROR)
        return

    if not token and error:
        error_msg = f'Error creating passkey for user ({user})!\nerror : {error}'
        logger.error(error_msg)
        with banner_container:
            st.error(error_msg, icon=config.ICON_ERROR)
        return

    if token:
        msg = f'Successfully registered user: {user.username}!'  # type: ignore
        logger.info(msg)
        with banner_container:
            st.success(msg, icon=config.ICON_SUCCESS)



r"""The client to use to interact with Bitwarden Passwordless."""

# Standard library
import logging
from typing import Any

# Third party
from pydantic import AnyHttpUrl, Field, PrivateAttr

# Local
from streamlit_passwordless import models

from . import backend

logger = logging.getLogger(__name__)


class BitwardenPasswordlessClient(models.BaseModel):
    r"""The client for interacting with Bitwarden Passwordless.

    Parameters
    ----------

    public_key : str
        The public key of the Bitwarden Passwordless application.

    private_key : str
        The private key of the Bitwarden Passwordless application.

    url : pydantic.AnyHttpUrl or str, default 'https://v4.passwordless.dev'
        The base url of the Bitwarden Passwordless application.

    register_config : streamlit_passwordless.BitwardenRegisterConfig
        The passkey configuration when registering a new user.
        If not specified the default configuration is used.
    """

    public_key: str
    private_key: str
    url: AnyHttpUrl = AnyHttpUrl('https://v4.passwordless.dev')
    register_config: backend.BitwardenRegisterConfig = Field(
        default_factory=backend.BitwardenRegisterConfig
    )
    _backend_client: backend.BackendClient = PrivateAttr()

    def model_post_init(self, __context: Any) -> None:
        r"""Setup the Bitwarden Passwordless backend client."""

        self._backend_client = backend._build_backend_client(
            private_key=self.private_key, url=str(self.url)
        )

    def create_register_token(self, user: models.User) -> str:
        r"""Create a register token to use for registering a passkey with the user's device.

        Parameters
        ----------
        user : streamlit_passwordless.User
            The user to register.

        Returns
        -------
        str
            The register token.

        Raises
        ------
        streamlit_passwordless.RegisterUserError
            If an error occurs while trying to create the register token
            using the Bitwarden Passwordless backend API.
        """

        return backend._create_register_token(
            client=self._backend_client,
            user=user,
            register_config=self.register_config,  # type: ignore
        )

    def verify_sign_in(self, token: str) -> backend.BitwardenPasswordlessVerifiedUser:
        r"""Verify the sign in token with the backend to complete the sign in process.

        Parameters
        ----------
        token : str
            The token to verify.

        Returns
        -------
        streamlit_passwordless.BitwardenPasswordlessVerifiedUser
            Details from Bitwarden Passwordless about the user that was signed in.

        Raises
        ------
        streamlit_passwordless.SignInTokenVerificationError
            If the `token` token cannot be verified successfully.

        streamlit_passwordless.StreamlitPasswordlessError
            If an instance of `backend.BitwardenPasswordlessVerifiedUser` cannot be
            successfully created.
        """

        return backend._verify_sign_in_token(client=self._backend_client, token=token)



r"""The data models of streamlit-passwordless."""

# Standard library
import uuid

# Third party
from pydantic import BaseModel as PydanticBaseModel
from pydantic import Field, ValidationError, field_validator

# Local
from . import exceptions


class BaseModel(PydanticBaseModel):
    r"""The BaseModel that all models inherit from."""

    def __init__(self, **kwargs) -> None:
        try:
            super().__init__(**kwargs)
        except ValidationError as e:
            raise exceptions.StreamlitPasswordlessError(str(e)) from None


class User(BaseModel):
    r"""A user within the streamlit-passwordless data model.

    Parameters
    ----------
    username : str
        The username of the user.

    user_id : str | None, default None
        The unique ID of the user which serves as the primary key in the database.
        If None it will be generated as a uuid.

    email : str | None, default None
        The email address of the user.

    displayname : str | None, default None
        A descriptive name of the user that is easy to understand for a human.

    aliases : tuple[str, ...] | str | None, default None
        Additional ID:s that can be used to identify the user when signing in.
        A string with aliases separated by semicolon ";" can be used to supply multiple
        aliases if tuple is not used.
    """

    username: str
    user_id: str | None = Field(default=None, validate_default=True)
    email: str | None = None
    displayname: str | None = None
    aliases: tuple[str, ...] | str | None = Field(default=None, validate_default=True)

    @field_validator('user_id')
    def generate_user_id(cls, v: str | None) -> str:
        r"""Generate a user ID if not supplied."""

        return str(uuid.uuid4()) if v is None else v

    @field_validator('aliases')
    def process_aliases(cls, aliases: tuple[str, ...] | str | None) -> tuple[str, ...] | None:
        r"""Convert multiple aliases in a string to a tuple by splitting on the semicolon ";"."""

        if isinstance(aliases, str):
            return tuple(v for a in aliases.split(';') if (v := a.strip()))
        else:
            return aliases



r"""The components' library contains the web components of streamlit_passwordless."""

# Local
from .config import ICON_ERROR, ICON_SUCCESS, ICON_WARNING, init_session_state
from .register_form import bitwarden_register_form
from .sign_in_form import bitwarden_sign_in_form

# The Public API

__all__ = [
    'ICON_ERROR',
    'ICON_SUCCESS',
    'ICON_WARNING',
    'init_session_state',
    'bitwarden_register_form',
    'bitwarden_sign_in_form',
]



r"""Components related to the frontend library of Bitwarden Passwordless."""

# Standard library
from pathlib import Path
from typing import Literal

# Third party
import streamlit as st
import streamlit.components.v1 as components

_RELEASE = True
_COMPONENT_NAME = 'bitwarden_passwordless'
_BUILD_DIR = Path(__file__).parent / 'frontend' / 'build'
_DEV_URL = 'http://localhost:3001'


if _RELEASE:
    _bitwarden_passwordless_func = components.declare_component(_COMPONENT_NAME, path=_BUILD_DIR)  # type: ignore
else:
    _bitwarden_passwordless_func = components.declare_component(name=_COMPONENT_NAME, url=_DEV_URL)


def register_button(
    register_token: str,
    public_key: str,
    credential_nickname: str,
    disabled: bool = False,
    label: str = 'Register',
    button_type: Literal['primary', 'secondary'] = 'primary',
    key: str | None = None,
) -> tuple[str, dict | None, bool]:
    r"""Render the register button, which starts the register process when clicked.

    The register process creates and registers a passkey with the user's device.

    The return value from the button is also saved to the session state with a key
    defined by the `key` parameter if `key` is not None.

    Parameters
    ----------
    register_token : str
        The registration token used to authorize the creation of a passkey on the user's device.

    public_key : str
        The public key of the Bitwarden Passwordless application.

    credential_nickname : str
        A nickname for the passkey credential being registered to use for easier identification
        of the device being registered.

    disabled : bool, default False
        If True the button will be disabled and if False the button will be clickable.

    label : str, default 'Register'
        The label of the button.

    button_type : Literal['primary', 'secondary'], default 'primary'
        The styling of the button. Emulates the `type` parameter of :func:`streamlit.button`.

    key : str or None, default None
        An optional key that uniquely identifies this component. If this is None, and the
        component's arguments are changed, the component will be re-mounted in the Streamlit
        frontend and lose its current state.

    Returns
    -------
    token : str
        The public key of the created passkey, which the user will use for future sign-in
        operations. This key is saved to the Bitwarden Passwordless database. An empty string
        is returned if the button has not been clicked.

    error : dict
        An error object containing information about if the registration process was successful
        or not. None is returned if no errors occurred during the register process or if the
        button has not been clicked.

    clicked : bool
        True if the button was clicked and False otherwise.
    """

    value = _bitwarden_passwordless_func(
        action='register',
        register_token=register_token,
        public_key=public_key,
        credential_nickname=credential_nickname,
        disabled=disabled,
        label=label,
        button_type=button_type,
        key=key,
    )

    if value is None:
        return '', None, False
    else:
        token, error, nr_clicks = value
        session_state_key = f'_{key}-nr-clicks'
        clicked = False if st.session_state.get(session_state_key, 0) == nr_clicks else True
        st.session_state[session_state_key] = nr_clicks

        return token, error, clicked


def sign_in_button(
    public_key: str,
    alias: str | None = None,
    with_discoverable: bool = True,
    with_autofill: bool = False,
    disabled: bool = False,
    label: str = 'Sign in',
    button_type: Literal['primary', 'secondary'] = 'primary',
    key: str | None = None,
) -> tuple[str, dict | None, bool]:
    r"""Render the sign in button, which starts the sign in process when clicked.

    Uses the Bitwarden Passwordless javascript frontend client.

    The return value from the button is also saved to the session state with a key
    defined by the `key` parameter if key is not None.

    Parameters
    ----------
    public_key : str
        The public key of the Bitwarden Passwordless application.

    alias : str or None, default None
        The alias of the user to sign in. If specified it will override the other sign in
        methods `with_discoverable`, and `with_autofill`.

    with_discoverable : bool, default True
        If True the browser's native UI prompt will be used to select the passkey to use for
        signing in. If False the sign in method is disabled. If True it will override the
        value of the `with_autofill` parameter. If `alias` is specified it will override this
        sign in method.

    with_autofill : bool, default False
        If True the browser's native autofill UI will be used to select the passkey to use for
        signing in. If False the sign in method is disabled. This method of signing in is
        overridden if `alias` is specified or `with_discoverable` is True.

    disabled : bool, default False
        If True the button will be disabled and if False the button will be clickable.

    label : str, default 'Register'
        The label of the button.

    button_type : Literal['primary', 'secondary'], default 'primary'
        The styling of the button. Emulates the `type` parameter of :func:`streamlit.button`.

    key : str or None, default None
        An optional key that uniquely identifies this component. If this is None, and the
        component's arguments are changed, the component will be re-mounted in the Streamlit
        frontend and lose its current state.

    Returns
    -------
    token : str
        The verification token to be used by the Bitwarden Passwordless backend to authenticate
        the sign in process. An empty string is returned if the button has not been clicked.

    error : dict | None
        An error object containing information if there was an error with the sign in process.
        None is returned if no errors occurred during the sign in process or if the button
        has not been clicked.

    clicked : bool
        True if the button was clicked and False otherwise.
    """

    value = _bitwarden_passwordless_func(
        action='sign_in',
        public_key=public_key,
        alias=alias,
        with_discoverable=with_discoverable,
        with_autofill=with_autofill,
        disabled=disabled,
        label=label,
        button_type=button_type,
        key=key,
    )

    if value is None:
        return '', None, False
    else:
        token, error, nr_clicks = value
        session_state_key = f'_{key}-nr-clicks'
        clicked = False if st.session_state.get(session_state_key, 0) == nr_clicks else True
        st.session_state[session_state_key] = nr_clicks

        return token, error, clicked



r"""The Bitwarden Passwordless package."""

# Local
from .backend import BitwardenPasswordlessVerifiedUser, BitwardenRegisterConfig
from .client import BitwardenPasswordlessClient
from .frontend import register_button, sign_in_button

# The Public API

__all__ = [
    'BitwardenPasswordlessVerifiedUser',
    'BitwardenRegisterConfig',
    'BitwardenPasswordlessClient',
    'register_button',
    'sign_in_button',
]
